@@ -4,21 +4,22 @@ import itertools
 
 from .helpers import SeededTest, select_by_precision
 from ..vartypes import continuous_types
-from ..model import Model, Point, Potential
+from ..model import Model, Point, Potential, Deterministic
 from ..blocking import DictToVarBijection, DictToArrayBijection, ArrayOrdering
 from ..distributions import (DensityDist, Categorical, Multinomial, VonMises, Dirichlet,
-                             MvStudentT, MvNormal, ZeroInflatedPoisson, GaussianRandomWalk,
+                             MvStudentT, MvNormal, MatrixNormal, ZeroInflatedPoisson,
                              ZeroInflatedNegativeBinomial, Constant, Poisson, Bernoulli, Beta,
-                             BetaBinomial, HalfStudentT, StudentT, Weibull, Pareto, NormalMixture,
+                             BetaBinomial, HalfStudentT, StudentT, Weibull, Pareto,
                              InverseGamma, Gamma, Cauchy, HalfCauchy, Lognormal, Laplace,
                              NegativeBinomial, Geometric, Exponential, ExGaussian, Normal,
                              Flat, LKJCorr, Wald, ChiSquared, HalfNormal, DiscreteUniform,
-                             Bound, Uniform, Triangular, Binomial, SkewNormal, DiscreteWeibull, Gumbel,
-                             Interpolated, ZeroInflatedBinomial)
+                             Bound, Uniform, Triangular, Binomial, SkewNormal, DiscreteWeibull,
+                             Gumbel, Logistic, LogitNormal, Interpolated, ZeroInflatedBinomial,
+                             HalfFlat, AR1, KroneckerNormal)
 from ..distributions import continuous
 from pymc3.theanof import floatX
 from numpy import array, inf, log, exp
-from numpy.testing import assert_almost_equal
+from numpy.testing import assert_almost_equal, assert_allclose, assert_equal
 import numpy.random as nr
 import numpy as np
 import pytest
@@ -26,9 +27,10 @@ import pytest
 from scipy import integrate
 import scipy.stats.distributions as sp
 import scipy.stats
+from scipy.special import logit
 import theano
 import theano.tensor as tt
-
+from ..math import kronecker
 
 def get_lkj_cases():
     """
@@ -75,11 +77,18 @@ class Domain(object):
             self.shape)
 
     def __mul__(self, other):
-        return Domain(
-            [v * other for v in self.vals],
-            self.dtype,
-            (self.lower * other, self.upper * other),
-            self.shape)
+        try:
+            return Domain(
+                [v * other for v in self.vals],
+                self.dtype,
+                (self.lower * other, self.upper * other),
+                self.shape)
+        except TypeError:
+            return Domain(
+                [v * other for v in self.vals],
+                self.dtype,
+                (self.lower, self.upper),
+                self.shape)
 
     def __neg__(self):
         return Domain(
@@ -248,6 +257,42 @@ def normal_logpdf_chol_upper(value, mu, chol):
     return normal_logpdf_cov(value, mu, np.dot(chol.T, chol)).sum()
 
 
+def matrix_normal_logpdf_cov(value, mu, rowcov, colcov):
+    return scipy.stats.matrix_normal.logpdf(value, mu, rowcov, colcov)
+
+
+def matrix_normal_logpdf_chol(value, mu, rowchol, colchol):
+    return matrix_normal_logpdf_cov(value, mu, np.dot(rowchol, rowchol.T),
+                                    np.dot(colchol, colchol.T))
+
+
+def kron_normal_logpdf_cov(value, mu, covs, sigma):
+    cov = kronecker(*covs).eval()
+    if sigma is not None:
+        cov += sigma**2 * np.eye(*cov.shape)
+    return scipy.stats.multivariate_normal.logpdf(value, mu, cov).sum()
+
+
+def kron_normal_logpdf_chol(value, mu, chols, sigma):
+    covs = [np.dot(chol, chol.T) for chol in chols]
+    return kron_normal_logpdf_cov(value, mu, covs, sigma=sigma)
+
+
+def kron_normal_logpdf_evd(value, mu, evds, sigma):
+    covs = []
+    for eigs, Q in evds:
+        try:
+            eigs = eigs.eval()
+        except AttributeError:
+            pass
+        try:
+            Q = Q.eval()
+        except AttributeError:
+            pass
+        covs.append(np.dot(Q, np.dot(np.diag(eigs), Q.T)))
+    return kron_normal_logpdf_cov(value, mu, covs, sigma)
+
+
 def betafn(a):
     return floatX(scipy.special.gammaln(a).sum(-1) - scipy.special.gammaln(a.sum(-1)))
 
@@ -274,17 +319,21 @@ def categorical_logpdf(value, p):
 
 def mvt_logpdf(value, nu, Sigma, mu=0):
     d = len(Sigma)
-    X = np.atleast_2d(value) - mu
-    Q = X.dot(np.linalg.inv(Sigma)).dot(X.T).sum()
-    log_det = np.log(np.linalg.det(Sigma))
+    dist = np.atleast_2d(value) - mu
+    chol = np.linalg.cholesky(Sigma)
+    trafo = np.linalg.solve(chol, dist.T).T
+    logdet = np.log(np.diag(chol)).sum()
 
-    log_pdf = scipy.special.gammaln(0.5 * (nu + d))
-    log_pdf -= 0.5 * (d * np.log(np.pi * nu) + log_det)
-    log_pdf -= scipy.special.gammaln(nu / 2.)
-    log_pdf -= 0.5 * (nu + d) * np.log(1 + Q / nu)
-    return log_pdf
+    lgamma = scipy.special.gammaln
+    norm = lgamma((nu + d) / 2.)  - 0.5 * d * np.log(nu * np.pi) - lgamma(nu / 2.)
+    logp = norm - logdet - (nu + d) / 2. * np.log1p((trafo * trafo).sum(-1) / nu)
+    return logp.sum()
 
 
+def AR1_logpdf(value, k, tau_e):
+    return (sp.norm(loc=0,scale=1/np.sqrt(tau_e)).logpdf(value[0]) +
+            sp.norm(loc=k*value[:-1],scale=1/np.sqrt(tau_e)).logpdf(value[1:]).sum())
+
 class Simplex(object):
     def __init__(self, n):
         self.vals = list(simplex_values(n))
@@ -353,12 +402,23 @@ def PdMatrixCholUpper(n):
         raise ValueError("n out of bounds")
 
 
+def RandomPdMatrix(n):
+    A = np.random.rand(n, n)
+    return np.dot(A, A.T) + n * np.identity(n)
+
+
 class TestMatchesScipy(SeededTest):
-    def pymc3_matches_scipy(self, pymc3_dist, domain, paramdomains, scipy_dist, decimal=None, extra_args={}):
+    def pymc3_matches_scipy(self, pymc3_dist, domain, paramdomains, scipy_dist,
+                            decimal=None, extra_args=None, scipy_args=None):
+        if extra_args is None:
+            extra_args = {}
+        if scipy_args is None:
+            scipy_args = {}
         model = build_model(pymc3_dist, domain, paramdomains, extra_args)
         value = model.named_vars['value']
 
         def logp(args):
+            args.update(scipy_args)
             return scipy_dist(**args)
         self.check_logp(model, value, domain, paramdomains, logp, decimal=decimal)
 
@@ -374,7 +434,7 @@ class TestMatchesScipy(SeededTest):
 
     def check_int_to_1(self, model, value, domain, paramdomains):
         pdf = model.fastfn(exp(model.logpt))
-        for pt in product(paramdomains, n_samples=100):
+        for pt in product(paramdomains, n_samples=10):
             pt = Point(pt, value=value.tag.test_value, model=model)
             bij = DictToVarBijection(value, (), pt)
             pdfx = bij.mapf(pdf)
@@ -409,9 +469,12 @@ class TestMatchesScipy(SeededTest):
             decimals = select_by_precision(float64=6, float32=4)
             assert_almost_equal(dlogp(pt), ndlogp(pt), decimal=decimals, err_msg=str(pt))
 
-    def checkd(self, distfam, valuedomain, vardomains, checks=None, extra_args={}):
+    def checkd(self, distfam, valuedomain, vardomains, checks=None, extra_args=None):
         if checks is None:
             checks = (self.check_int_to_1, self.check_dlogp)
+
+        if extra_args is None:
+            extra_args = {}
         m = build_model(distfam, valuedomain, vardomains, extra_args=extra_args)
         for check in checks:
             check(m, m.named_vars['value'], valuedomain, vardomains)
@@ -426,7 +489,6 @@ class TestMatchesScipy(SeededTest):
             Triangular, Runif, {'lower': -Rplusunif, 'c': Runif, 'upper': Rplusunif},
             lambda value, c, lower, upper: sp.triang.logpdf(value, c-lower, lower, upper-lower))
 
-
     def test_bound_normal(self):
         PositiveNormal = Bound(Normal, lower=0.)
         self.pymc3_matches_scipy(PositiveNormal, Rplus, {'mu': Rplus, 'sd': Rplus},
@@ -442,6 +504,16 @@ class TestMatchesScipy(SeededTest):
 
     def test_flat(self):
         self.pymc3_matches_scipy(Flat, Runif, {}, lambda value: 0)
+        with Model():
+            x = Flat('a')
+            assert_allclose(x.tag.test_value, 0)
+
+    def test_half_flat(self):
+        self.pymc3_matches_scipy(HalfFlat, Rplus, {}, lambda value: 0)
+        with Model():
+            x = HalfFlat('a', shape=2)
+            assert_allclose(x.tag.test_value, 1)
+            assert x.tag.test_value.shape == (2,)
 
     def test_normal(self):
         self.pymc3_matches_scipy(Normal, R, {'mu': R, 'sd': Rplus},
@@ -556,10 +628,10 @@ class TestMatchesScipy(SeededTest):
                                  scipy_exponweib_sucks,
                                  )
 
-    def test_student_tpos(self):
-        # TODO: this actually shouldn't pass
-        self.pymc3_matches_scipy(HalfStudentT, Rplus, {'nu': Rplus, 'mu': R, 'lam': Rplus},
-                                 lambda value, nu, mu, lam: sp.t.logpdf(value, nu, mu, lam**-.5))
+    def test_half_studentt(self):
+        # this is only testing for nu=1 (halfcauchy)
+        self.pymc3_matches_scipy(HalfStudentT, Rplus, {'sd': Rplus},
+                                 lambda value, sd: sp.halfcauchy.logpdf(value, 0, sd))
 
     def test_skew_normal(self):
         self.pymc3_matches_scipy(SkewNormal, R, {'mu': R, 'sd': Rplusbig, 'alpha': R},
@@ -569,14 +641,19 @@ class TestMatchesScipy(SeededTest):
         self.pymc3_matches_scipy(Binomial, Nat, {'n': NatSmall, 'p': Unit},
                                  lambda value, n, p: sp.binom.logpmf(value, n, p))
 
-    @pytest.mark.xfail(condition=(theano.config.floatX == "float32"), reason="Fails on float32")  # Too lazy to propagate decimal parameter through the whole chain of deps
+    # Too lazy to propagate decimal parameter through the whole chain of deps
+    @pytest.mark.xfail(condition=(theano.config.floatX == "float32"), reason="Fails on float32")
     def test_beta_binomial(self):
         self.checkd(BetaBinomial, Nat, {'alpha': Rplus, 'beta': Rplus, 'n': NatSmall})
 
     def test_bernoulli(self):
+        self.pymc3_matches_scipy(
+            Bernoulli, Bool, {'logit_p': R},
+            lambda value, logit_p: sp.bernoulli.logpmf(value, scipy.special.expit(logit_p)))
         self.pymc3_matches_scipy(Bernoulli, Bool, {'p': Unit},
                                  lambda value, p: sp.bernoulli.logpmf(value, p))
 
+
     def test_discrete_weibull(self):
         self.pymc3_matches_scipy(DiscreteWeibull, Nat,
                 {'q': Unit, 'beta': Rplusdunif}, discrete_weibull_logpmf)
@@ -597,16 +674,19 @@ class TestMatchesScipy(SeededTest):
         self.pymc3_matches_scipy(Constant, I, {'c': I},
                                  lambda value, c: np.log(c == value))
 
-    @pytest.mark.xfail(condition=(theano.config.floatX == "float32"), reason="Fails on float32")  # Too lazy to propagate decimal parameter through the whole chain of deps
+    # Too lazy to propagate decimal parameter through the whole chain of deps
+    @pytest.mark.xfail(condition=(theano.config.floatX == "float32"), reason="Fails on float32")
     def test_zeroinflatedpoisson(self):
         self.checkd(ZeroInflatedPoisson, Nat, {'theta': Rplus, 'psi': Unit})
 
-    @pytest.mark.xfail(condition=(theano.config.floatX == "float32"), reason="Fails on float32")  # Too lazy to propagate decimal parameter through the whole chain of deps
+    # Too lazy to propagate decimal parameter through the whole chain of deps
+    @pytest.mark.xfail(condition=(theano.config.floatX == "float32"), reason="Fails on float32")
     def test_zeroinflatednegativebinomial(self):
         self.checkd(ZeroInflatedNegativeBinomial, Nat,
                     {'mu': Rplusbig, 'alpha': Rplusbig, 'psi': Unit})
 
-    @pytest.mark.xfail(condition=(theano.config.floatX == "float32"), reason="Fails on float32")  # Too lazy to propagate decimal parameter through the whole chain of deps
+    # Too lazy to propagate decimal parameter through the whole chain of deps
+    @pytest.mark.xfail(condition=(theano.config.floatX == "float32"), reason="Fails on float32")
     def test_zeroinflatedbinomial(self):
         self.checkd(ZeroInflatedBinomial, Nat,
                     {'n': NatSmall, 'p': Unit, 'psi': Unit})
@@ -671,11 +751,90 @@ class TestMatchesScipy(SeededTest):
             with pytest.raises(ValueError):
                 x = MvNormal('x', mu=np.zeros(3), cov=np.eye(3), tau=np.eye(3), shape=3)
 
+    @pytest.mark.parametrize('n', [1, 2, 3])
+    def test_matrixnormal(self, n):
+        mat_scale = 1e3  # To reduce logp magnitude
+        mean_scale = .1
+        self.pymc3_matches_scipy(MatrixNormal, RealMatrix(n, n),
+                                 {'mu': RealMatrix(n, n)*mean_scale,
+                                  'rowcov': PdMatrix(n)*mat_scale,
+                                  'colcov': PdMatrix(n)*mat_scale},
+                                 matrix_normal_logpdf_cov)
+        self.pymc3_matches_scipy(MatrixNormal, RealMatrix(2, n),
+                                 {'mu': RealMatrix(2, n)*mean_scale,
+                                  'rowcov': PdMatrix(2)*mat_scale,
+                                  'colcov': PdMatrix(n)*mat_scale},
+                                 matrix_normal_logpdf_cov)
+        self.pymc3_matches_scipy(MatrixNormal, RealMatrix(3, n),
+                                 {'mu': RealMatrix(3, n)*mean_scale,
+                                  'rowchol': PdMatrixChol(3)*mat_scale,
+                                  'colchol': PdMatrixChol(n)*mat_scale},
+                                 matrix_normal_logpdf_chol,
+                                 decimal=select_by_precision(float64=6, float32=-1))
+        self.pymc3_matches_scipy(MatrixNormal, RealMatrix(n, 3),
+                                 {'mu': RealMatrix(n, 3)*mean_scale,
+                                  'rowchol': PdMatrixChol(n)*mat_scale,
+                                  'colchol': PdMatrixChol(3)*mat_scale},
+                                 matrix_normal_logpdf_chol,
+                                 decimal=select_by_precision(float64=6, float32=0))
+
+    @pytest.mark.parametrize('n', [2, 3])
+    @pytest.mark.parametrize('m', [3])
+    @pytest.mark.parametrize('sigma', [None, 1.0])
+    def test_kroneckernormal(self, n, m, sigma):
+        np.random.seed(5)
+        N = n*m
+        covs = [RandomPdMatrix(n), RandomPdMatrix(m)]
+        chols = list(map(np.linalg.cholesky, covs))
+        evds = list(map(np.linalg.eigh, covs))
+        dom = Domain([np.random.randn(N)*0.1], edges=(None, None), shape=N)
+        mu = Domain([np.random.randn(N)*0.1], edges=(None, None), shape=N)
+
+        std_args = {'mu': mu}
+        cov_args = {'covs': covs}
+        chol_args = {'chols': chols}
+        evd_args = {'evds': evds}
+        if sigma is not None and sigma != 0:
+            std_args['sigma'] = Domain([sigma], edges=(None, None))
+        else:
+            for args in [cov_args, chol_args, evd_args]:
+                args['sigma'] = sigma
+
+        self.pymc3_matches_scipy(
+             KroneckerNormal, dom, std_args, kron_normal_logpdf_cov,
+             extra_args=cov_args, scipy_args=cov_args)
+        self.pymc3_matches_scipy(
+             KroneckerNormal, dom, std_args, kron_normal_logpdf_chol,
+             extra_args=chol_args, scipy_args=chol_args)
+        self.pymc3_matches_scipy(
+             KroneckerNormal, dom, std_args, kron_normal_logpdf_evd,
+             extra_args=evd_args, scipy_args=evd_args)
+
+        dom = Domain([np.random.randn(2, N)*0.1], edges=(None, None), shape=(2, N))
+
+        self.pymc3_matches_scipy(
+             KroneckerNormal, dom, std_args, kron_normal_logpdf_cov,
+             extra_args=cov_args, scipy_args=cov_args)
+        self.pymc3_matches_scipy(
+             KroneckerNormal, dom, std_args, kron_normal_logpdf_chol,
+             extra_args=chol_args, scipy_args=chol_args)
+        self.pymc3_matches_scipy(
+             KroneckerNormal, dom, std_args, kron_normal_logpdf_evd,
+             extra_args=evd_args, scipy_args=evd_args)
+
     @pytest.mark.parametrize('n', [1, 2])
     def test_mvt(self, n):
         self.pymc3_matches_scipy(MvStudentT, Vector(R, n),
                                  {'nu': Rplus, 'Sigma': PdMatrix(n), 'mu': Vector(R, n)},
                                  mvt_logpdf)
+        self.pymc3_matches_scipy(MvStudentT, RealMatrix(2, n),
+                                 {'nu': Rplus, 'Sigma': PdMatrix(n), 'mu': Vector(R, n)},
+                                 mvt_logpdf)
+
+    @pytest.mark.parametrize('n',[2,3,4])
+    def test_AR1(self, n):
+        self.pymc3_matches_scipy(AR1, Vector(R, n), {'k': Unit, 'tau_e': Rplus}, AR1_logpdf)
+
 
     @pytest.mark.parametrize('n', [2, 3])
     def test_wishart(self, n):
@@ -711,6 +870,51 @@ class TestMatchesScipy(SeededTest):
         self.pymc3_matches_scipy(Multinomial, Vector(Nat, n), {'p': Simplex(n), 'n': Nat},
                                  multinomial_logpdf)
 
+    @pytest.mark.parametrize('p,n', [
+        [[.25, .25, .25, .25], 1],
+        [[.3, .6, .05, .05], 2],
+        [[.3, .6, .05, .05], 10],
+    ])
+    def test_multinomial_mode(self, p, n):
+        _p = np.array(p)
+        with Model() as model:
+            m = Multinomial('m', n, _p, _p.shape)
+        assert_allclose(m.distribution.mode.eval().sum(), n)
+        _p = np.array([p, p])
+        with Model() as model:
+            m = Multinomial('m', n, _p, _p.shape)
+        assert_allclose(m.distribution.mode.eval().sum(axis=-1), n)
+
+    @pytest.mark.parametrize('p, shape, n', [
+        [[.25, .25, .25, .25], 4, 2],
+        [[.25, .25, .25, .25], (1, 4), 3],
+        # 3: expect to fail
+        # [[.25, .25, .25, .25], (10, 4)],
+        [[.25, .25, .25, .25], (10, 1, 4), 5],
+        # 5: expect to fail
+        # [[[.25, .25, .25, .25]], (2, 4), [7, 11]],
+        [[[.25, .25, .25, .25],
+         [.25, .25, .25, .25]], (2, 4), 13],
+        [[[.25, .25, .25, .25],
+         [.25, .25, .25, .25]], (2, 4), [17, 19]],
+        [[[.25, .25, .25, .25],
+         [.25, .25, .25, .25]], (1, 2, 4), [23, 29]],
+        [[[.25, .25, .25, .25],
+         [.25, .25, .25, .25]], (10, 2, 4), [31, 37]],
+    ])
+    def test_multinomial_random(self, p, shape, n):
+        p = np.asarray(p)
+        with Model() as model:
+            m = Multinomial('m', n=n, p=p, shape=shape)
+        m.random()
+
+    def test_multinomial_mode_with_shape(self):
+        n = [1, 10]
+        p = np.asarray([[.25,.25,.25,.25], [.26, .26, .26, .22]])
+        with Model() as model:
+            m = Multinomial('m', n=n, p=p, shape=(2, 4))
+        assert_allclose(m.distribution.mode.eval().sum(axis=-1), n)
+
     def test_multinomial_vec(self):
         vals = np.array([[2,4,4], [3,3,4]])
         p = np.array([0.2, 0.3, 0.5])
@@ -722,6 +926,14 @@ class TestMatchesScipy(SeededTest):
         with Model() as model_many:
             Multinomial('m', n=n, p=p, shape=vals.shape)
 
+        assert_almost_equal(scipy.stats.multinomial.logpmf(vals, n, p),
+                            np.asarray([model_single.fastlogp({'m': val}) for val in vals]),
+                            decimal=4)
+
+        assert_almost_equal(scipy.stats.multinomial.logpmf(vals, n, p),
+                            model_many.free_RVs[0].logp_elemwise({'m': vals}).squeeze(),
+                            decimal=4)
+
         assert_almost_equal(sum([model_single.fastlogp({'m': val}) for val in vals]),
                             model_many.fastlogp({'m': vals}),
                             decimal=4)
@@ -808,6 +1020,7 @@ class TestMatchesScipy(SeededTest):
         pt = {'eg': value}
         assert_almost_equal(model.fastlogp(pt), logp, decimal=select_by_precision(float64=6, float32=2), err_msg=str(pt))
 
+    @pytest.mark.xfail(condition=(theano.config.floatX == "float32"), reason="Fails on float32")
     def test_vonmises(self):
         self.pymc3_matches_scipy(
             VonMises, R, {'mu': Circ, 'kappa': Rplus},
@@ -818,6 +1031,17 @@ class TestMatchesScipy(SeededTest):
             return floatX(sp.gumbel_r.logpdf(value, loc=mu, scale=beta))
         self.pymc3_matches_scipy(Gumbel, R, {'mu': R, 'beta': Rplusbig}, gumbel)
 
+    def test_logistic(self):
+        self.pymc3_matches_scipy(Logistic, R, {'mu': R, 's': Rplus},
+                                 lambda value, mu, s: sp.logistic.logpdf(value, mu, s),
+                                 decimal=select_by_precision(float64=6, float32=1))
+
+    def test_logitnormal(self):
+        self.pymc3_matches_scipy(LogitNormal, Unit, {'mu': R, 'sd': Rplus},
+                                 lambda value, mu, sd: (sp.norm.logpdf(logit(value), mu, sd)
+                                                        - (np.log(value) + np.log1p(-value))),
+                                 decimal=select_by_precision(float64=6, float32=1))
+
     def test_multidimensional_beta_construction(self):
         with Model():
             Beta('beta', alpha=1., beta=1., shape=(10, 20))
@@ -850,27 +1074,117 @@ class TestMatchesScipy(SeededTest):
                 self.pymc3_matches_scipy(TestedInterpolated, R, {}, ref_pdf)
 
 
-def test_repr_latex_():
-    with Model():
-        x0 = Binomial('Discrete', p=.5, n=10)
-        x1 = Normal('Continuous', mu=0., sd=1.)
-        x2 = GaussianRandomWalk('Timeseries', mu=x1, sd=1., shape=2)
-        x3 = MvStudentT('Multivariate', nu=5, mu=x2, Sigma=np.diag(np.ones(2)), shape=2)
-        x4 = NormalMixture('Mixture', w=np.array([.5, .5]), mu=x3, sd=x0)
+def test_bound():
+    np.random.seed(42)
+    UnboundNormal = Bound(Normal)
+    dist = UnboundNormal.dist(mu=0, sd=1)
+    assert dist.transform is None
+    assert dist.default() == 0.
+    assert isinstance(dist.random(), np.ndarray)
 
-    assert x0._repr_latex_() == '$Discrete \\sim \\text{Binomial}' \
-                                '(\\mathit{n}=10, \\mathit{p}=0.5)$'
-    assert x1._repr_latex_() == '$Continuous \\sim \\text{Normal}' \
-                                '(\\mathit{mu}=0.0, \\mathit{sd}=1.0)$'
-    assert x2._repr_latex_() == '$Timeseries \\sim \\text' \
-                                '{GaussianRandomWalk}(\\mathit{mu}=Continuous, ' \
-                                '\\mathit{sd}=1.0)$'
-    assert x3._repr_latex_() == '$Multivariate \\sim \\text{MvStudentT}' \
-                                '(\\mathit{nu}=5, \\mathit{mu}=Timeseries, ' \
-                                '\\mathit{Sigma}=array)$'
-    assert x4._repr_latex_() == '$Mixture \\sim \\text{NormalMixture}' \
-                                '(\\mathit{w}=array, \\mathit{mu}=Multivariate, ' \
-                                '\\mathit{sigma}=f(Discrete))$'
+    LowerNormal = Bound(Normal, lower=1)
+    dist = LowerNormal.dist(mu=0, sd=1)
+    assert dist.logp(0).eval() == -np.inf
+    assert dist.default() > 1
+    assert dist.transform is not None
+    assert np.all(dist.random() > 1)
+
+    UpperNormal = Bound(Normal, upper=-1)
+    dist = UpperNormal.dist(mu=0, sd=1)
+    assert dist.logp(-0.5).eval() == -np.inf
+    assert dist.default() < -1
+    assert dist.transform is not None
+    assert np.all(dist.random() < -1)
+
+    ArrayNormal = Bound(Normal, lower=[1, 2], upper=[2, 3])
+    dist = ArrayNormal.dist(mu=0, sd=1, shape=2)
+    assert_equal(dist.logp([0.5, 3.5]).eval(), -np.array([np.inf, np.inf]))
+    assert_equal(dist.default(), np.array([1.5, 2.5]))
+    assert dist.transform is not None
+    with pytest.raises(ValueError) as err:
+        dist.random()
+    err.match('Drawing samples from distributions with array-valued')
+
+    with Model():
+        a = ArrayNormal('c', shape=2)
+        assert_equal(a.tag.test_value, np.array([1.5, 2.5]))
+
+    lower = tt.vector('lower')
+    lower.tag.test_value = np.array([1, 2]).astype(theano.config.floatX)
+    upper = 3
+    ArrayNormal = Bound(Normal, lower=lower, upper=upper)
+    dist = ArrayNormal.dist(mu=0, sd=1, shape=2)
+    logp = dist.logp([0.5, 3.5]).eval({lower: lower.tag.test_value})
+    assert_equal(logp, -np.array([np.inf, np.inf]))
+    assert_equal(dist.default(), np.array([2, 2.5]))
+    assert dist.transform is not None
+
+    with Model():
+        a = ArrayNormal('c', shape=2)
+        assert_equal(a.tag.test_value, np.array([2, 2.5]))
+
+    rand = Bound(Binomial, lower=10).dist(n=20, p=0.3).random()
+    assert rand.dtype in [np.int16, np.int32, np.int64]
+    assert rand >= 10
+
+    rand = Bound(Binomial, upper=10).dist(n=20, p=0.8).random()
+    assert rand.dtype in [np.int16, np.int32, np.int64]
+    assert rand <= 10
+
+    rand = Bound(Binomial, lower=5, upper=8).dist(n=10, p=0.6).random()
+    assert rand.dtype in [np.int16, np.int32, np.int64]
+    assert rand >= 5 and rand <= 8
+
+
+class TestLatex(object):
+
+    def setup_class(self):
+        # True parameter values
+        alpha, sigma = 1, 1
+        beta = [1, 2.5]
+
+        # Size of dataset
+        size = 100
+
+        # Predictor variable
+        X = np.random.normal(size=(size, 2)).dot(np.array([[1, 0], [0, 0.2]]))
+
+        # Simulate outcome variable
+        Y = alpha + X.dot(beta) + np.random.randn(size)*sigma
+        with Model() as self.model:
+            # Priors for unknown model parameters
+            alpha = Normal('alpha', mu=0, sd=10)
+            b = Normal('beta', mu=0, sd=10, shape=(2,), observed=beta)
+            sigma = HalfNormal('sigma', sd=1)
+
+            # Expected value of outcome
+            mu = Deterministic('mu', alpha + tt.dot(X, b))
+
+            # Likelihood (sampling distribution) of observations
+            Y_obs = Normal('Y_obs', mu=mu, sd=sigma, observed=Y)
+        self.distributions = [alpha, sigma, mu, b, Y_obs]
+        self.expected = (
+            r'$\text{alpha} \sim \text{Normal}(\mathit{mu}=0,~\mathit{sd}=10.0)$',
+            r'$\text{sigma} \sim \text{HalfNormal}(\mathit{sd}=1.0)$',
+            r'$\text{mu} \sim \text{Deterministic}(\text{alpha},~\text{Constant},~\text{beta})$',
+            r'$\text{beta} \sim \text{Normal}(\mathit{mu}=0,~\mathit{sd}=10.0)$',
+            r'$\text{Y_obs} \sim \text{Normal}(\mathit{mu}=\text{mu},~\mathit{sd}=f(\text{sigma}))$'
+        )
+
+    def test__repr_latex_(self):
+        for distribution, tex in zip(self.distributions, self.expected):
+            assert distribution._repr_latex_() == tex
+
+        model_tex = self.model._repr_latex_()
+
+        for tex in self.expected:  # make sure each variable is in the model
+            for segment in tex.strip('$').split(r'\sim'):
+                assert segment in model_tex
+
+    def test___latex__(self):
+        for distribution, tex in zip(self.distributions, self.expected):
+            assert distribution._repr_latex_() == distribution.__latex__()
+        assert self.model._repr_latex_() == self.model.__latex__()
 
 
 def test_discrete_trafo():
